# Copyright 2020 - 2025 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Serialization and Deserialization Errors."""


class SerdeError(Exception):
    """Serialization and Deserialization Error."""
